namespace X3DTK {
namespace MESH {

template<class MData, class VData, class EData, class FData, bool readOnly>
FacesDrawerCoreVisitor<MData, VData, EData, FData, readOnly>::FacesDrawerCoreVisitor()
{
  // Enter and leave functions.
  defineEnterFunction(&FacesDrawerCoreVisitor::enterTransform);
  defineEnterFunction(&FacesDrawerCoreVisitor::enterMesh);
  defineWalkOnFunction(&FacesDrawerCoreVisitor::walkOnShape);
  defineLeaveFunction(&FacesDrawerCoreVisitor::leaveTransform);
  
  // StateVariables assignation.
  stateVariables = GraphTraversal::getInstanceOf<FacesDrawerStateVariables<MData, VData, EData, FData, readOnly> >();
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void FacesDrawerCoreVisitor<MData, VData, EData, FData, readOnly>::enterTransform(Transform *T) const
{
  glMatrixMode(GL_MODELVIEW);
  stateVariables->pushMatrix(T->getTransform());
  glPushMatrix();
  glMultMatrixf(T->getTransform().toFloat16());
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void FacesDrawerCoreVisitor<MData, VData, EData, FData, readOnly>::leaveTransform(Transform *) const
{
  glMatrixMode(GL_MODELVIEW);
  stateVariables->popMatrix();
  glPopMatrix();
}

template<class MData, class VData, class EData, class FData, bool readOnly>
bool FacesDrawerCoreVisitor<MData, VData, EData, FData, readOnly>::walkOnShape(Shape *, SFAbstractNode child) const
{
  MFNode selectedNodes = stateVariables->getSelectedNodes();
  return (selectedNodes.empty() || (std::find(selectedNodes.begin(), selectedNodes.end(), child) != selectedNodes.end()));
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void FacesDrawerCoreVisitor<MData, VData, EData, FData, readOnly>::enterMesh(Mesh<MData, VData, EData, FData, readOnly> *M) const
{
  glEnable(GL_LIGHTING);
  
  stateVariables->storeMesh(M);
  int meshId = stateVariables->getCurrentMeshId();

  if (stateVariables->getSelectionMode())
    glPushName(meshId);

  bool colorPerVertex = true;
  if (MData::template find<MESH::MeshColorData>())
    colorPerVertex = M->template ogetData<MESH::MeshColorData>().getColorPerVertex();
  
  bool normalPerVertex = true;
  if (MData::template find<MESH::MeshNormalData>())
    normalPerVertex = M->template ogetData<MESH::MeshNormalData>().getNormalPerVertex();
  
  const typename Mesh<MData, VData, EData, FData, readOnly>::MFFace &faces = M->getFaces();

  if (stateVariables->getRenderingMode() != REAL)
  {
    SFNode N = 0;
    if (stateVariables->getRenderingMode() == IFS)
      N = M;
    else
      N = M->getVertex();

    static const unsigned int cr = 101;
    static const unsigned int cg = 117;
    static const unsigned int cb = 131;
  
    static const float icr = 1.0f/(2.0f*static_cast<float>(cr));
    static const float icg = 1.0f/(2.0f*static_cast<float>(cg));
    static const float icb = 1.0f/(2.0f*static_cast<float>(cb));
  
    float meshAddr = icr*static_cast<float>(reinterpret_cast<unsigned int>(N) % cr);
    float meshAddg = icg*static_cast<float>(reinterpret_cast<unsigned int>(N) % cg);
    float meshAddb = icb*static_cast<float>(reinterpret_cast<unsigned int>(N) % cb);
  
    glColor3f(meshAddr + 0.5f, meshAddg + 0.5f, meshAddb + 0.5f);
  }
  
  for (typename Mesh<MData, VData, EData, FData, readOnly>::MFFace::const_iterator f = faces.begin(), end = faces.end() ; f != end; ++f)
  {
    if (FData::template find<MESH::FaceNormalData>() && !normalPerVertex)
    {
      SFVec3f normal = (*f)->template ogetData<MESH::FaceNormalData>().getNormal();
      glNormal3fv(normal.f_data());
    }

    if (FData::template find<MESH::FaceColorData>() && !colorPerVertex && stateVariables->getRenderingMode() == REAL)
    {
      SFColorRGBA color = (*f)->template ogetData<MESH::FaceColorData>().getColor();
      glColor4f(color.r, color.g, color.b, color.a);
    }
        
    if (stateVariables->getSelectionMode())
      glPushName((*f)->getIndex());

    glBegin(GL_POLYGON);
    const typename SFFace<FData, VData, EData, readOnly>::MFEdge &edges = (*f)->getEdges();
    for (typename SFFace<FData, VData, EData, readOnly>::MFEdge::const_iterator e = edges.begin(), end = edges.end() ; e != end; ++e)
    {
      if (VData::template find<MESH::VertexNormalData>() && normalPerVertex)
      {
        SFVec3f normal = (*e)->getFromVertex()->template ogetData<MESH::VertexNormalData>().getNormalOfFace(*f);
        glNormal3fv(normal.f_data());
      }
      if (VData::template find<MESH::VertexColorData>() && colorPerVertex && stateVariables->getRenderingMode() == REAL)
      {
        SFColorRGBA color = (*e)->getFromVertex()->template ogetData<MESH::VertexColorData>().getColorOfFace(*f);
        glColor4f(color.r, color.g, color.b, color.a);
      }
      
      SFPoint3f point = (*e)->getFromVertex()->template ogetData<MESH::VertexPointData>().getPoint();
      glVertex3fv(point.f_data());
    }
    glEnd();

    if (stateVariables->getSelectionMode())
      glPopName();
  }

  if (stateVariables->getSelectionMode())
    glPopName();
    
  glDisable(GL_LIGHTING);  
}


}
}  
