/// This file is part of the X3DToolKit library
/// Copyright (C) 2002-2004  Yannick Le Goc (legoc@imag.fr)

/// http://artis.imag.fr/Members/Yannick.Legoc/X3D/

/// This library is free software; you can redistribute it and/or
/// modify it under the terms of the GNU Lesser General Public
/// License as published by the Free Software Foundation; either
/// version 2.1 of the License, or (at your option) any later version.

/// This library is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/// Lesser General Public License for more details.

/// You should have received a copy of the GNU Lesser General Public
/// License along with this library; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#ifndef X3DTK_MESH_MESH_H
#define X3DTK_MESH_MESH_H

#include "MESH_X3DMeshNode.h"
#include "MESH_SFVertex.h"
#include "MESH_Vertex.h"
#include "MESH_SFDirectEdge.h"
#include "MESH_SFIndirectEdge.h"
#include "MESH_SFFace.h"
#include "MESH_MeshData.h"

#include <iostream>
#include <typeinfo>

namespace X3DTK {
namespace MESH {

/*! \brief Template class declared in <b><X3DTK/MESH/scenegraph.h></b> defining a Mesh,
 *  the most important structure of the MESH scene graph. 
 *
 *  It provides a set of interconnected vertices, oriented-edges and faces.
 *
 *  \ingroup MeshSceneGraph
 */


template<class MData, class VData, class EData, class FData, bool readOnly = true>
class Mesh : public X3DMeshNode
{
public:
  /// Defines the Vertex type for the Mesh.
  typedef SFVertex<VData, EData, FData, readOnly> SFVertex;
  /// Defines the MVertex type for the Mesh.
  typedef MeshMFVertex<VData, EData, FData, readOnly> MFVertex;
  /// Defines the Edge type for the Mesh.  
  typedef SFEdge<EData, FData, VData, readOnly> SFEdge;
  /// Defines the MEdge type for the Mesh.  
  typedef MeshMFEdge<EData, FData, VData, readOnly> MFEdge;
  /// Defines the Face type for the Mesh.  
  typedef SFFace<FData, VData, EData, readOnly> SFFace;
  /// Defines the MFace type for the Mesh.  
  typedef MeshMFFace<FData, VData, EData, readOnly> MFFace;
  
  /// Constructor.
  Mesh();
  /// Clones the node.
  virtual SFNode clone() const;
  /// Virtual destructor.
  virtual ~Mesh(); 
   
  /// Sets the Vertex attribute.
  void setVertex(Vertex<VData, EData, FData, readOnly> *vertexSet);
  /// Gets the children list.
  virtual MFAbstractNode getChildrenList() const;
  
  /// Adds the child N. Returns FALSE if the child already exists or N is not a 
  /// valid child.   
  bool addChild(const SFAbstractNode &N);
  /// Sets the child N. Returns FALSE if N is not a 
  /// valid child. If the child already exists then N becomes the new child.
  bool setChild(const SFAbstractNode &N);
  /// Removes the child if exists. Returns TRUE if succeeds.
  bool removeChild(const SFAbstractNode &N);
 
  /// Get the edge from - to.
  SFEdge *getEdge(SFVertex *from, SFVertex *to);
  /// Gets the Vertices that can be shared with other Mesh.
  const MFVertex &getVertices() const;
    
  /// Gets the edges.
  inline const MFEdge &getEdges() const;
  /// Gets the faces.
  inline const MFFace &getFaces() const;
  /// Gets the data.  
  inline MData &data();
  /// Gets the data.  
  inline const MData &data() const;
  /// Gets the data F from MData which must be of type clist<tlist<T...> >.
  template<class F>  
  F &getData();
  /// Gets optionally the data F from MData which must be of type clist<tlist<T...> >.
  template<class F>  
  F &ogetData();
    
  /// Creates Face at the end. The vertices present in coordIndex must already be created,
  /// otherwise, it returns a NULL Face.
  SFFace *createFace(const MFInt32 &coordIndex, bool ccw = true);
  /// Creates a new face at position i.
  SFFace *createFace(unsigned int i, const MFInt32 &coordIndex, bool ccw = true);

  /// Removes face.
  void removeFace(SFFace *face);
  /// Removes face at index i.
  void removeFace(unsigned int i);
  
protected:
  /// Copy constructor.
  Mesh(const Mesh &N);
  
private:
  MFEdge _edges;
  MFFace _faces;
  MData _data;
  
  /// Creates a new edge.
  SFEdge *createEdge(SFVertex *from, SFVertex *to);
};

/// Mesh for the default X3DNode nodes containing all the geometry and topology 
/// informations of an X3D scene.
typedef Mesh<MeshData, VertexData, EdgeData, FaceData, true> DefMesh;
typedef SFVertex<VertexData, EdgeData, FaceData, true> DefSFVertex;
typedef SFEdge<EdgeData, FaceData, VertexData, true> DefSFEdge;
typedef SFFace<FaceData, VertexData, EdgeData, true> DefSFFace;

}
}

#include "MESH_Mesh.inl"

#endif
