/// This file is part of the X3DToolKit library
/// Copyright (C) 2002-2004  Yannick Le Goc (legoc@imag.fr)

/// http://artis.imag.fr/Members/Yannick.Legoc/X3D/

/// This library is free software; you can redistribute it and/or
/// modify it under the terms of the GNU Lesser General Public
/// License as published by the Free Software Foundation; either
/// version 2.1 of the License, or (at your option) any later version.

/// This library is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/// Lesser General Public License for more details.

/// You should have received a copy of the GNU Lesser General Public
/// License along with this library; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#ifndef X3DTK_MESH_MESHDATA_H
#define X3DTK_MESH_MESHDATA_H

#include "X3DTypes.h"
#include "TypeList.h"
#include "MESH_SFFace.h"

#include <map>

namespace X3DTK {
namespace MESH {

/*! \brief Structure declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D vertex point data.
 *
 *  \ingroup MeshSceneGraph
 */

struct VertexPointData
{
  /// Sets the point.
  void setPoint(const SFPoint3f &point) {_point = point;};
  /// Gets the point.
  inline const SFPoint3f &getPoint() const {return _point;};
  
private:
  SFPoint3f _point;  
};

/*! \brief Structure declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D vertex normal data.
 *
 *  \ingroup MeshSceneGraph
 */

struct VertexNormalData
{
  /// Sets the normal of face. 
  template<class FData, class VData, class EData, bool readOnly>
  void setNormalOfFace(const SFVec3f &normal, SFFace<FData, VData, EData, readOnly> *face);
  /// Gets the normal of face.
  template<class FData, class VData, class EData, bool readOnly>
  const SFVec3f &getNormalOfFace(SFFace<FData, VData, EData, readOnly> *face) const;
  
  /// Gets the normal map.
  const std::map<void *, unsigned short> &getMap() const {return _normalMap;};
  /// Gets the normal array.
  const MFVec3f &getNormalArray() const {return _normalArray;};
  
private:
  MFVec3f _normalArray; 
  std::map<void *, unsigned short> _normalMap;
};

/*! \brief Structure declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D vertex color data.
 *
 *  \ingroup MeshSceneGraph
 */

struct VertexColorData
{
  /// Sets the normal of face.
  template<class FData, class VData, class EData, bool readOnly>
  void setColorOfFace(const SFColorRGBA &color, SFFace<FData, VData, EData, readOnly> *face);
  /// Sets the normal of face.
  template<class FData, class VData, class EData, bool readOnly>
  const SFColorRGBA &getColorOfFace(SFFace<FData, VData, EData, readOnly> *face) const;
  
private:
  MFColorRGBA _colorArray;
  std::map<void *, unsigned short> _colorMap;
};

/*! \brief Structure declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D vertex texture coordinate data.
 *
 *  \ingroup MeshSceneGraph
 */

struct VertexTexCoordData
{
  /// Sets the normal of face.
  template<class FData, class VData, class EData, bool readOnly>
  void setTexCoordOfFace(const SFVec2f &texCoord, SFFace<FData, VData, EData, readOnly> *face);
  /// Sets the normal of face.
  template<class FData, class VData, class EData, bool readOnly>
  const SFVec2f &getTexCoordOfFace(SFFace<FData, VData, EData, readOnly> *face) const;
   
private:
  MFVec2f _texCoordArray;
  std::map<void *, unsigned short> _texCoordMap;
};

/*! \brief Structure declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D face normal data.
 *
 *  \ingroup MeshSceneGraph
 */

struct FaceNormalData
{
  /// Sets the normal.
  void setNormal(const SFVec3f &normal) {_normal = normal;};
  /// Gets the normal.
  inline const SFVec3f &getNormal() const {return _normal;};
  
private:
  SFVec3f _normal;  
};

/*! \brief Structure declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D face color data.
 *
 *  \ingroup MeshSceneGraph
 */

struct FaceColorData
{
  /// Sets the normal.
  void setColor(const SFColorRGBA &color) {_color = color;};
  /// Gets the normal.
  inline const SFColorRGBA &getColor() const {return _color;};
  
private:
  SFColorRGBA _color;  
};

/*! \brief Structure declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D mesh normal data.
 *
 *  \ingroup MeshSceneGraph
 */

struct MeshNormalData
{
  /// Sets the boolean normal, true if vertex has normal.
  void setNormal(bool value) {_normal = value;};
  /// Sets the boolean normalPerVertex.
  void setNormalPerVertex(bool value) {_normalPerVertex = value;};
  
  /// Returns true if the Mesh has normals.
  inline bool hasNormal() const {return _normal;};
  /// Gets the boolean faceNormal.
  inline bool getNormalPerVertex() const {return _normalPerVertex;};
  
private:
  bool _normal;
  bool _normalPerVertex;
};

/*! \brief Structure declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D mesh color data.
 *
 *  \ingroup MeshSceneGraph
 */

struct MeshColorData
{
  /// Sets the boolean Color, true if vertex has Color.
  void setColor(bool value) {_color = value;};
  /// Sets the boolean colorPerVertex.
  void setColorPerVertex(bool value) {_colorPerVertex = value;};
  /// Sets the boolean RGBA.
  void setRGBA(bool value) {_rgba = value;};
  
  /// Returns true if the Mesh has Colors.
  inline bool hasColor() const {return _color;};
  /// Gets the boolean faceColor.
  inline bool getColorPerVertex() const {return _colorPerVertex;};
  /// Gets the boolean RGBA.
  inline bool getRGBA() const {return _rgba;};
  
private:
  bool _color;
  bool _colorPerVertex;
  bool _rgba;
};

/*! \brief Structure declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D mesh texture coordinate data.
 *
 *  \ingroup MeshSceneGraph
 */

struct MeshTexCoordData
{
  /// Sets the boolean TexCoord, true if vertex has TexCoord.
  void setTexCoord(bool value) {_texCoord = value;};
  /// Returns true if the Mesh has TexCoords.
  inline bool hasTexCoord() const {return _texCoord;};
  
private:
  bool _texCoord;
};

/*! \brief Type declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D vertex datas.
 *
 *  \ingroup MeshSceneGraph
 */

typedef clist<tlist<VertexPointData, tlist<VertexNormalData, tlist<VertexColorData, tlist<VertexTexCoordData> > > > > VertexData;

/*! \brief Type declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D edge datas.
 *
 *  \ingroup MeshSceneGraph
 */

typedef nil EdgeData;

/*! \brief Structure declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D face data.
 *
 *  \ingroup MeshSceneGraph
 */

typedef clist<tlist<FaceNormalData, tlist<FaceColorData> > > FaceData;

/*! \brief Structure declared in <b><X3DTK/MESH/scenegraph.h></b> defining the default X3D mesh data.
 *
 *  \ingroup MeshSceneGraph
 */

typedef clist<tlist<MeshNormalData, tlist<MeshColorData, tlist<MeshTexCoordData> > > > MeshData;

}
}

#include "MESH_MeshData.inl"

#endif
