/// This file is part of the X3DToolKit library
/// Copyright (C) 2002-2004  Yannick Le Goc (legoc@imag.fr)

/// http://artis.imag.fr/Members/Yannick.Legoc/X3D/

/// This library is free software; you can redistribute it and/or
/// modify it under the terms of the GNU Lesser General Public
/// License as published by the Free Software Foundation; either
/// version 2.1 of the License, or (at your option) any later version.

/// This library is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/// Lesser General Public License for more details.

/// You should have received a copy of the GNU Lesser General Public
/// License along with this library; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#ifndef MESH_SCENEGRAPHTYPES_H
#define MESH_SCENEGRAPHTYPES_H

#include "Platform.h"

#include <vector>
#include <map>
#include <set>
#include <list>
#include <iostream>

namespace X3DTK {
namespace MESH {

template<class A, class B>
struct ConstMapIterator : public std::map<A, B>::const_iterator
{
  ConstMapIterator(const typename std::map<A, B>::const_iterator &it) : std::map<A, B>::const_iterator(it) {}
  
  const B &operator*() const 
  {
    return (this->std::map<A, B>::const_iterator::operator*()).second;
  }  
};

template<class A, class B>
struct NonConstMapIterator : public std::map<A, B>::iterator
{
  NonConstMapIterator(const typename std::map<A, B>::iterator &it) : std::map<A, B>::iterator(it) {}
  
  B &operator*() const 
  {
    return (this->std::map<A, B>::iterator::operator*()).second;
  }  
};

class X3DNode;

/*! \brief Type declared in <b><X3DTK/MESH/scenegraph.h></b> providing a pointer to X3DNode.
 *
 *  \ingroup MeshSceneGraph
 */

typedef X3DNode *SFNode;

/*! \brief Class declared in <b><X3DTK/MESH/scenegraph.h></b> providing a multiple pointer to X3DNode values
 *  deriving from std::list<X3DNode *>.
 *
 *  \ingroup MeshSceneGraph
 */
 
typedef std::list<SFNode> MFNode;

template<class VData, class EData, class FData, bool readOnly> class SFVertex;
template<class EData, class FData, class VData, bool readOnly> class SFEdge;
template<class FData, class VData, class EData, bool readOnly> class SFFace;

/*! \brief Class declared in <b><X3DTK/MESH/scenegraph.h></b> providing a multiple SFVertex values
 *
 *  \ingroup MeshSceneGraph
 */

template<class VData, class EData, class FData, bool readOnly = true>
class EntityMFVertex : public std::vector<SFVertex<VData, EData, FData, readOnly> *>
{
public:
  /// Constructor.
  EntityMFVertex();
  /// Constructor from a vector.
  explicit EntityMFVertex(const std::vector<SFVertex<VData, EData, FData, readOnly> *> &vector);
};

template<class VData, class EData, class FData>
class EntityMFVertex<VData, EData, FData, false> : public std::list<SFVertex<VData, EData, FData, false> *>
{
public:
  /// Constructor.
  EntityMFVertex();
  /// Constructor from a vector.
  explicit EntityMFVertex(const std::vector<SFVertex<VData, EData, FData, false> *> &vector);
};

/*! \brief Class declared in <b><X3DTK/MESH/scenegraph.h></b> providing a multiple SFVertex values
 *
 *  \ingroup MeshSceneGraph
 */

template<class VData, class EData, class FData, bool readOnly = true>
class MeshMFVertex : public std::vector<SFVertex<VData, EData, FData, readOnly> *>
{
public:
  /// Constructor.
  MeshMFVertex();
  /// Constructor from a vector.
  explicit MeshMFVertex(const std::vector<SFVertex<VData, EData, FData, readOnly> *> &vector);

  /// Inserts element x at index x.first of the std::vector. Is provided to have common interface with std::map.
  std::pair<typename std::vector<SFVertex<VData, EData, FData, readOnly> *>::iterator, bool> insert(const std::pair<unsigned int, SFVertex<VData, EData, FData, readOnly> *> &x)
  {
    if (x.first >= size())
      resize(x.first + 1, 0);
  
    typename std::vector<SFVertex<VData, EData, FData, readOnly> *>::iterator it = begin();
    it += x.first;
    *it = x.second;
    
    return std::pair<typename std::vector<SFVertex<VData, EData, FData, readOnly> *>::iterator, bool>(it, true);
  }

  /// Finds the element. Is provided for having a common interface with std::set and std::map.
  typename std::vector<SFVertex<VData, EData, FData, readOnly> *>::iterator find(SFVertex<VData, EData, FData, readOnly> *const &element) const
  {
    return std::find(begin(), end(), element);
  }

};

template<class VData, class EData, class FData>
class MeshMFVertex<VData, EData, FData, false> : public std::map<unsigned int, SFVertex<VData, EData, FData, false> *>
{
public:
  typedef NonConstMapIterator<unsigned int, SFVertex<VData, EData, FData, false> *> iterator;
  typedef ConstMapIterator<unsigned int, SFVertex<VData, EData, FData, false> *> const_iterator;

  /// Constructor.
  MeshMFVertex();
  /// Constructor from a vector.
  explicit MeshMFVertex(const std::vector<SFVertex<VData, EData, FData, false> *> &vector);
  
  /// operator[].
  SFVertex<VData, EData, FData, false> *const &operator[](const unsigned int &n) const 
  {
    return (*std::map<unsigned int, SFVertex<VData, EData, FData, false> *>::find(n)).second;
  }
  
  /// Finds the element. Is provided for having a common interface with std::set and std::map.
  typename std::map<unsigned int, SFVertex<VData, EData, FData, false> *>::iterator find(SFVertex<VData, EData, FData, false> *const &element)
  {
    typename std::map<unsigned int, SFVertex<VData, EData, FData, false> *>::iterator it = std::map<unsigned int, SFVertex<VData, EData, FData, false> *>::find(element->getIndex());
    if ((*it).second == element)
      return it;
    return end();  
  }
  
  /// Finds the element. Is provided for having a common interface with std::set and std::map.
  typename std::map<unsigned int, SFVertex<VData, EData, FData, false> *>::const_iterator find(SFVertex<VData, EData, FData, false> *const &element) const
  {
    typename std::map<unsigned int, SFVertex<VData, EData, FData, false> *>::const_iterator it = std::map<unsigned int, SFVertex<VData, EData, FData, false> *>::find(element->getIndex());
    if ((*it).second == element)
      return it;
    return end();  
  }
};


/*! \brief Class declared in <b><X3DTK/MESH/scenegraph.h></b> providing a multiple SFEdge values
 *
 *  \ingroup MeshSceneGraph
 */

template<class EData, class FData, class VData, bool readOnly = true>
class EntityMFEdge : public std::vector<SFEdge<EData, FData, VData, readOnly> *>
{
public:
  /// Constructor.
  EntityMFEdge();
  /// Constructor from a vector.
  explicit EntityMFEdge(const std::vector<SFEdge<EData, FData, VData, readOnly> *> &vector, bool ccw);
};

template<class EData, class FData, class VData>
class EntityMFEdge<EData, FData, VData, false> : public std::list<SFEdge<EData, FData, VData, false> *>
{
public:
  /// Constructor.
  EntityMFEdge();
  /// Constructor from a vector.
  explicit EntityMFEdge(const std::vector<SFEdge<EData, FData, VData, false> *> &vector, bool ccw);
};

/*! \brief Class declared in <b><X3DTK/MESH/scenegraph.h></b> providing a multiple SFEdge values
 *
 *  \ingroup MeshSceneGraph
 */

template<class EData, class FData, class VData, bool readOnly = true>
class MeshMFEdge : public std::vector<SFEdge<EData, FData, VData, readOnly> *>
{
public:
  /// Constructor.
  MeshMFEdge();
  /// Constructor from a vector.
  explicit MeshMFEdge(const std::vector<SFEdge<EData, FData, VData, readOnly> *> &vector, bool ccw);

  /// Inserts element x at the end of the std::vector. Is equivalent to push_back and provided to have common interface with std::map.
  std::pair<typename std::vector<SFEdge<EData, FData, VData, readOnly> *>::iterator, bool> insert(const typename std::vector<SFEdge<EData, FData, VData, readOnly> *>::value_type &x)
  {
    push_back(x);
    return std::pair<typename std::vector<SFEdge<EData, FData, VData, readOnly> *>::iterator, bool>(end(), true);
  }

  /// Finds the element. Is provided for having a common interface with std::set and std::map.
  typename std::vector<SFEdge<EData, FData, VData, readOnly> *>::iterator find(SFEdge<EData, FData, VData, readOnly> *const &element) const
  {
    return std::find(begin(), end(), element);
  }
};

template<class EData, class FData, class VData>
class MeshMFEdge<EData, FData, VData, false> : public std::set<SFEdge<EData, FData, VData, false> *>
{
public:
  /// Constructor.
  MeshMFEdge();
  /// Constructor from a vector.
  explicit MeshMFEdge(const std::vector<SFEdge<EData, FData, VData, false> *> &vector, bool ccw);

  /// operator[].
  SFEdge<EData, FData, VData, false> *const &operator[](const unsigned int &n) const
  {
    typename std::set<SFEdge<EData, FData, VData, false> *>::const_iterator it = begin();
    for (unsigned int i = 0; i < n; ++i)
      ++it;
    return *it;  
  }
};

/*! \brief Class declared in <b><X3DTK/MESH/scenegraph.h></b> providing a multiple SFFace values
 *
 *  \ingroup MeshSceneGraph
 */

template<class FData, class VData, class EData, bool readOnly = true>
class EntityMFFace : public std::vector<SFFace<FData, VData, EData, readOnly> *>
{
public:
  /// Constructor.
  EntityMFFace();
  /// Constructor from a vector.
  explicit EntityMFFace(const std::vector<SFFace<FData, VData, EData, readOnly> *> &vector);
};

template<class FData, class VData, class EData>
class EntityMFFace<FData, VData, EData, false> : public std::list<SFFace<FData, VData, EData, false> *>
{
public:
  /// Constructor.
  EntityMFFace();
  /// Constructor from a vector.
  explicit EntityMFFace(const std::vector<SFFace<FData, VData, EData, false> *> &vector);
};

/*! \brief Class declared in <b><X3DTK/MESH/scenegraph.h></b> providing a multiple SFFace values
 *
 *  \ingroup MeshSceneGraph
 */

template<class FData, class VData, class EData, bool readOnly = true>
class MeshMFFace : public std::vector<SFFace<FData, VData, EData, readOnly> *>
{
public:
  /// Constructor.
  MeshMFFace();
  /// Constructor from a vector.
  explicit MeshMFFace(const std::vector<SFFace<FData, VData, EData, readOnly> *> &vector);

  /// Inserts element x at the end of the std::vector. Is equivalent to push_back and provided to have common interface with std::map.
  std::pair<typename std::vector<SFFace<FData, VData, EData, readOnly> *>::iterator, bool> insert(const std::pair<unsigned int, SFFace<FData, VData, EData, readOnly> *> &x)
  {
    if (x.first >= size())
      resize(x.first + 1, 0);
  
    typename std::vector<SFFace<FData, VData, EData, readOnly> *>::iterator it = begin();
    it += x.first;
    *it = x.second;
    
    return std::pair<typename std::vector<SFFace<FData, VData, EData, readOnly> *>::iterator, bool>(it, true);
  }

  /// Finds the element. Is provided for having a common interface with std::set and std::map.
  typename std::vector<SFFace<FData, VData, EData, readOnly> *>::iterator find(SFFace<FData, VData, EData, readOnly> *const &element)
  {
    return std::find(begin(), end(), element);
  }
  /// Finds the element. Is provided for having a common interface with std::set and std::map.
  typename std::vector<SFFace<FData, VData, EData, readOnly> *>::const_iterator find(SFFace<FData, VData, EData, readOnly> *const &element) const
  {
    return std::find(begin(), end(), element);
  }
};

template<class FData, class VData, class EData>
class MeshMFFace<FData, VData, EData, false> : public std::map<unsigned int, SFFace<FData, VData, EData, false> *>
{
public:
  typedef NonConstMapIterator<unsigned int, SFFace<FData, VData, EData, false> *> iterator;
  typedef ConstMapIterator<unsigned int, SFFace<FData, VData, EData, false> *> const_iterator;

  /// Constructor.
  MeshMFFace();
  /// Constructor from a vector.
  explicit MeshMFFace(const std::vector<SFFace<FData, VData, EData, false> *> &vector);

  /// operator [].
  SFFace<FData, VData, EData, false> * const &operator[](const unsigned int &n) const 
  {
    return (*std::map<unsigned int, SFFace<FData, VData, EData, false> *>::find(n)).second;
  }

  /// Finds the element. Is provided for having a common interface with std::set and std::map.
  typename std::map<unsigned int, SFFace<FData, VData, EData, false> *>::iterator find(SFFace<FData, VData, EData, false> *const &element)
  {
    typename std::map<unsigned int, SFFace<FData, VData, EData, false> *>::iterator it = std::map<unsigned int, SFFace<FData, VData, EData, false> *>::find(element->getIndex());
    if ((*it).second == element)  
      return it;
    return end();
  }
  /// Finds the element. Is provided for having a common interface with std::set and std::map.
  typename std::map<unsigned int, SFFace<FData, VData, EData, false> *>::const_iterator find(SFFace<FData, VData, EData, false> *const &element) const
  {
    typename std::map<unsigned int, SFFace<FData, VData, EData, false> *>::const_iterator it = std::map<unsigned int, SFFace<FData, VData, EData, false> *>::find(element->getIndex());
    if ((*it).second == element)  
      return it;
    return end(); 
  }
};

}
}

#include "MESH_SceneGraphTypes.inl"

#endif
