namespace X3DTK {
namespace MESH {

template<class MData, class VData, class EData, class FData, bool readOnly>
X3DBuilderCoreVisitor<MData, VData, EData, FData, readOnly>::X3DBuilderCoreVisitor()
{
  defineEnterFunction(&X3DBuilderCoreVisitor::enterScene);
  defineEnterFunction(&X3DBuilderCoreVisitor::enterTransform);
  defineEnterFunction(&X3DBuilderCoreVisitor::enterSwitch);
  defineEnterFunction(&X3DBuilderCoreVisitor::enterShape);
  defineEnterFunction(&X3DBuilderCoreVisitor::enterMesh);
  defineEnterFunction(&X3DBuilderCoreVisitor::enterVertex);
  defineLeaveFunction(&X3DBuilderCoreVisitor::leaveX3DNode);

  stateVariables = GraphTraversal::getInstanceOf<X3DBuilderStateVariables<MData, VData, EData, FData, readOnly> >();
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void X3DBuilderCoreVisitor<MData, VData, EData, FData, readOnly>::enterScene(Scene *S) const
{
  X3D::Scene *NS = dynamic_cast<X3D::Scene *>(stateVariables->getNode(S));
  if (NS == 0)
  {
    NS = new X3D::Scene();
    NS->setFileName(S->getFileName());
    // Updating Scene.
    X3D::MFNode x3dChildren = S->getX3DChildren();
    for (X3D::MFNode::const_iterator it = x3dChildren.begin(); it != x3dChildren.end(); ++it)
      NS->addChild(stateVariables->clone(*it));
  
    stateVariables->addCoupleNode(S, NS);
  }
  stateVariables->pushNode(NS); 
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void X3DBuilderCoreVisitor<MData, VData, EData, FData, readOnly>::enterTransform(Transform *T) const
{
  X3D::Transform *NT = dynamic_cast<X3D::Transform *>(stateVariables->getNode(T));
  if (NT == 0)
  {
    NT = new X3D::Transform();
    NT->setName(T->getName());
    
    // Updating Transform.
    NT->setCenter(T->getCenter());
    NT->setRotation(T->getRotation());
    NT->setScale(T->getScale());
    NT->setScaleOrientation(T->getScaleOrientation());
    NT->setTranslation(T->getTranslation());
  
    X3D::MFNode x3dChildren = T->getX3DChildren();
    for (X3D::MFNode::const_iterator it = x3dChildren.begin(); it != x3dChildren.end(); ++it)
      NT->addChild(stateVariables->clone(*it));
  
    stateVariables->addCoupleNode(T, NT);
  }
  stateVariables->pushNode(NT); 
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void X3DBuilderCoreVisitor<MData, VData, EData, FData, readOnly>::enterSwitch(Switch *S) const
{
  X3D::Switch *NS = dynamic_cast<X3D::Switch *>(stateVariables->getNode(S));
  if (NS == 0)
  {
    NS = new X3D::Switch();
    NS->setName(S->getName());
    
    // Updating Switch.
    NS->setWhichChoice(S->getWhichChoice());
  
    X3D::MFNode x3dChildren = S->getX3DChildren();
    for (X3D::MFNode::const_iterator it = x3dChildren.begin(); it != x3dChildren.end(); ++it)
      NS->addChild(stateVariables->clone(*it));
  
    stateVariables->addCoupleNode(S, NS);
  }
  stateVariables->pushNode(NS); 
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void X3DBuilderCoreVisitor<MData, VData, EData, FData, readOnly>::enterShape(Shape *S) const
{
  X3D::Shape *NS = dynamic_cast<X3D::Shape *>(stateVariables->getNode(S));
  if (NS == 0)
  {
    NS = new X3D::Shape();
    NS->setName(S->getName());

    NS->setAppearance(dynamic_cast<X3D::X3DAppearanceNode *>(stateVariables->clone(S->getAppearance())));
    
    stateVariables->addCoupleNode(S, NS);
  }
  stateVariables->pushNode(NS);   
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void X3DBuilderCoreVisitor<MData, VData, EData, FData, readOnly>::enterMesh(Mesh<MData, VData, EData, FData, readOnly> *M) const
{
  X3D::IndexedFaceSet *NI = dynamic_cast<X3D::IndexedFaceSet *>(stateVariables->getNode(M));
  if (NI == 0)
  {
    NI = new X3D::IndexedFaceSet();
    NI->setName(M->getName());
  
    // Updating IndexedFaceSet.
    // setting the parameters.
    NI->setCcw(true);
    
    if (MData::template find<MESH::MeshNormalData>())
      NI->setNormalPerVertex(M->template ogetData<MESH::MeshNormalData>().getNormalPerVertex());

    if (MData::template find<MESH::MeshColorData>())      
      NI->setColorPerVertex(M->template ogetData<MESH::MeshColorData>().getColorPerVertex());
  
    // Faces
    MFInt32 coordIndexArray;
    const typename Mesh<MData, VData, EData, FData, readOnly>::MFFace &mf = M->getFaces(); 
  
    // Filling the coordIndex.
    for (typename Mesh<MData, VData, EData, FData, readOnly>::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
    {
      const typename SFFace<FData, VData, EData, readOnly>::MFEdge &me = (*itFace)->getEdges();
      for (typename SFFace<FData, VData, EData, readOnly>::MFEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
        coordIndexArray.push_back((*itEdge)->getFromVertex()->getIndex());
      coordIndexArray.push_back(-1);
    }

    NI->setCoordIndex(coordIndexArray);
  
    ///////////////////////////////////////////////////////////////////////////////////
    // normals
    
    if (MData::template find<MESH::MeshNormalData>()
     && M->template ogetData<MESH::MeshNormalData>().hasNormal())
    {
      X3D::Normal *N = new X3D::Normal();
      NI->setNormal(N);
    
      MFVec3f normalArray;

      if (VData::template find<MESH::VertexNormalData>()
       && M->template ogetData<MESH::MeshNormalData>().getNormalPerVertex())
      {
        // iterating the faces to create the normalIndexArray.
        MFInt32 normalIndexArray;
        
        // Using a temporary map to keep the computed normals and find faster whether the
        // normal is in the NormalArray or not.
        
        std::map<SFVertex<VData, EData, FData, readOnly> *, std::list<unsigned int> > _vertexMap; 
    
        for (typename Mesh<MData, VData, EData, FData, readOnly>::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
        {
          const typename SFFace<FData, VData, EData, readOnly>::MFEdge &me = (*itFace)->getEdges();
          
          for (typename SFFace<FData, VData, EData, readOnly>::MFEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
          {
            SFVertex<VData, EData, FData, readOnly> *V = (*itEdge)->getFromVertex();
            const SFVec3f &normal = V->template ogetData<MESH::VertexNormalData>().getNormalOfFace(*itFace);
            
            // finding the normal among the list of normals for V, already present in NormalArray
            std::list<unsigned int>::const_iterator itNormalIndex = _vertexMap[V].begin();
            while (itNormalIndex != _vertexMap[V].end())
            {
              if (normalArray[*itNormalIndex] == normal)
              {
                normalIndexArray.push_back(*itNormalIndex);
                break;
              }
              ++itNormalIndex;
            }
            
            // if not found, adding it to NormalArray
            if (itNormalIndex == _vertexMap[V].end())
            {
              _vertexMap[V].push_back(normalArray.size());
              normalIndexArray.push_back(normalArray.size());
              normalArray.push_back(normal);
            }
          }  
          normalIndexArray.push_back(-1);
        }
        NI->setNormalIndex(normalIndexArray);
      }
      
      if (FData::template find<MESH::FaceNormalData>()
       && !M->template ogetData<MESH::MeshNormalData>().getNormalPerVertex())
      {
        for (typename Mesh<MData, VData, EData, FData, readOnly>::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
          normalArray.push_back((*itFace)->template ogetData<MESH::FaceNormalData>().getNormal());
      }
      N->setVector(normalArray);
    }
  
    //////////////////////////////////////////////////////////////////////////////////// 
    // colors

    if (MData::template find<MESH::MeshColorData>()
     && M->template ogetData<MESH::MeshColorData>().hasColor()
     && M->template ogetData<MESH::MeshColorData>().getRGBA())
    {
      // creating a new Node.
      X3D::ColorRGBA *C = new X3D::ColorRGBA();
      NI->setColor(C);

      MFColorRGBA colorArray;

      if (VData::template find<MESH::VertexColorData>()
       && M->template ogetData<MESH::MeshColorData>().getColorPerVertex())
      {
        // iterating the faces to create the colorIndexArray.
        MFInt32 colorIndexArray;
    
        // Using a temporary map to keep the computed colors and find faster whether the
        // color is in the ColorArray or not.
        
        std::map<SFVertex<VData, EData, FData, readOnly> *, std::list<unsigned int> > _vertexMap; 
            
        for (typename Mesh<MData, VData, EData, FData, readOnly>::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
        {
          const typename SFFace<FData, VData, EData, readOnly>::MFEdge &me = (*itFace)->getEdges();
          for (typename SFFace<FData, VData, EData, readOnly>::MFEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
          {
            SFVertex<VData, EData, FData, readOnly> *V = (*itEdge)->getFromVertex();
            const SFColorRGBA &color = V->template ogetData<MESH::VertexColorData>().getColorOfFace(*itFace);

            // finding the color among the list of colors for V, already present in ColorArray
            std::list<unsigned int>::const_iterator itColorIndex = _vertexMap[V].begin();
            while (itColorIndex != _vertexMap[V].end())
            {
              if (colorArray[*itColorIndex] == color)
              {
                colorIndexArray.push_back(*itColorIndex);
                break;
              }
              ++itColorIndex;
            }
            
            // if not found, adding it to NormalArray
            if (itColorIndex == _vertexMap[V].end())
            {
              _vertexMap[V].push_back(colorArray.size());
              colorIndexArray.push_back(colorArray.size());
              colorArray.push_back(color);
            }
          }  
          colorIndexArray.push_back(-1);
        }
        NI->setColorIndex(colorIndexArray);
      }
      
      if (FData::template find<MESH::FaceColorData>()
       && !M->template ogetData<MESH::MeshColorData>().getColorPerVertex())
      {
        for (typename Mesh<MData, VData, EData, FData, readOnly>::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
          colorArray.push_back((*itFace)->template ogetData<MESH::FaceColorData>().getColor());
      }
      C->setColor(colorArray);
    }

    if (MData::template find<MESH::MeshColorData>()
     && M->template ogetData<MESH::MeshColorData>().hasColor()
     && !M->template ogetData<MESH::MeshColorData>().getRGBA())
    {
      // creating a new Node.
      X3D::Color *C = new X3D::Color();
      NI->setColor(C);

      MFColor colorArray;

      if (VData::template find<MESH::VertexColorData>()
       && M->template ogetData<MESH::MeshColorData>().getColorPerVertex())
      {
        // iterating the faces to create the colorIndexArray.
        MFInt32 colorIndexArray;
    
        // Using a temporary map to keep the computed colors and find faster whether the
        // color is in the ColorArray or not.
        
        std::map<SFVertex<VData, EData, FData, readOnly> *, std::list<unsigned int> > _vertexMap; 
            
        for (typename Mesh<MData, VData, EData, FData, readOnly>::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
        {
          const typename SFFace<FData, VData, EData, readOnly>::MFEdge &me = (*itFace)->getEdges();
          for (typename SFFace<FData, VData, EData, readOnly>::MFEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
          {
            SFVertex<VData, EData, FData, readOnly> *V = (*itEdge)->getFromVertex();

            const SFColorRGBA &colorRGBA = V->template ogetData<MESH::VertexColorData>().getColorOfFace(*itFace);
            SFColor color(colorRGBA.r, colorRGBA.g, colorRGBA.b);

            // finding the color among the list of colors for V, already present in ColorArray
            std::list<unsigned int>::const_iterator itColorIndex = _vertexMap[V].begin();
            while (itColorIndex != _vertexMap[V].end())
            {
              if (colorArray[*itColorIndex] == color)
              {
                colorIndexArray.push_back(*itColorIndex);
                break;
              }
              ++itColorIndex;
            }
            
            // if not found, adding it to NormalArray
            if (itColorIndex == _vertexMap[V].end())
            {
              _vertexMap[V].push_back(colorArray.size());
              colorIndexArray.push_back(colorArray.size());
              colorArray.push_back(color);
            }
          }  
          colorIndexArray.push_back(-1);
        }
        NI->setColorIndex(colorIndexArray);
      }
      
      if (FData::template find<MESH::FaceColorData>()
       && !M->template ogetData<MESH::MeshColorData>().getColorPerVertex())
      {
        for (typename Mesh<MData, VData, EData, FData, readOnly>::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
        {
          const SFColorRGBA &colorRGBA = (*itFace)->template ogetData<MESH::FaceColorData>().getColor();
          SFColor color(colorRGBA.r, colorRGBA.g, colorRGBA.b);
          colorArray.push_back(color);
        }  
      }
      C->setColor(colorArray);
    }


    //////////////////////////////////////////////////////////////////////////////////// 
    // texCoords

    if (MData::template find<MESH::MeshTexCoordData>()
     && VData::template find<MESH::VertexTexCoordData>()
     && M->template ogetData<MESH::MeshTexCoordData>().hasTexCoord())
    {
      X3D::TextureCoordinate *T = new X3D::TextureCoordinate();
      NI->setTexCoord(T);
  
      MFVec2f texCoordArray;
      // iterating the faces to create the normalIndexArray.
      MFInt32 texCoordIndexArray;
  
      // Using a temporary map to keep the computed texCoords and find faster whether the
      // texCoord is in the texCoordArray or not.
        
      std::map<SFVertex<VData, EData, FData, readOnly> *, std::list<unsigned int> > _vertexMap; 
  
      for (typename Mesh<MData, VData, EData, FData, readOnly>::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
      {
        const typename SFFace<FData, VData, EData, readOnly>::MFEdge &me = (*itFace)->getEdges();
        for (typename SFFace<FData, VData, EData, readOnly>::MFEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
        {
          SFVertex<VData, EData, FData, readOnly> *V = (*itEdge)->getFromVertex();
          const SFVec2f &texCoord = V->template ogetData<MESH::VertexTexCoordData>().getTexCoordOfFace(*itFace);
        
          // finding the color among the list of colors for V, already present in TexCoordArray
          std::list<unsigned int>::const_iterator itTexCoordIndex = _vertexMap[V].begin();
          while (itTexCoordIndex != _vertexMap[V].end())
          {
            if (texCoordArray[*itTexCoordIndex] == texCoord)
            {
              texCoordIndexArray.push_back(*itTexCoordIndex);
              break;
            }
            ++itTexCoordIndex;
          }
            
          // if not found, adding it to NormalArray
          if (itTexCoordIndex == _vertexMap[V].end())
          {
            _vertexMap[V].push_back(texCoordArray.size());
            texCoordIndexArray.push_back(texCoordArray.size());
            texCoordArray.push_back(texCoord);
          }
        }  
        texCoordIndexArray.push_back(-1);
      }
      NI->setTexCoordIndex(texCoordIndexArray);
      T->setPoint(texCoordArray);
    }
    stateVariables->addCoupleNode(M, NI);
  }
  stateVariables->pushNode(NI);
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void X3DBuilderCoreVisitor<MData, VData, EData, FData, readOnly>::enterVertex(Vertex<VData, EData, FData, readOnly> *V) const
{
  X3D::Coordinate *NC = dynamic_cast<X3D::Coordinate *>(stateVariables->getNode(V));
  if (NC == 0)
  {
    NC = new X3D::Coordinate();
    NC->setName(V->getName());
    
    if (VData::template find<MESH::VertexPointData>())
    {
      // Updating Coordinate.
      const typename Vertex<VData, EData, FData, readOnly>::MFVertex &mv = V->getVertices();
      MFVec3f point;
    
      for (typename Vertex<VData, EData, FData, readOnly>::MFVertex::const_iterator it = mv.begin(); it != mv.end(); ++it)
         point.push_back(SFVec3f((*it)->template ogetData<MESH::VertexPointData>().getPoint()));
    
      NC->setPoint(point);
    }
    
    stateVariables->addCoupleNode(V, NC);
  }
  stateVariables->pushNode(NC); 
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void X3DBuilderCoreVisitor<MData, VData, EData, FData, readOnly>::leaveX3DNode(X3DNode *) const
{
  stateVariables->popNode();
}

}
}
