/// This file is part of the X3DToolKit library
/// Copyright (C) 2002-2004  Yannick Le Goc (legoc@imag.fr)

/// http://artis.imag.fr/Members/Yannick.Legoc/X3D/

/// This library is free software; you can redistribute it and/or
/// modify it under the terms of the GNU Lesser General Public
/// License as published by the Free Software Foundation; either
/// version 2.1 of the License, or (at your option) any later version.

/// This library is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/// Lesser General Public License for more details.

/// You should have received a copy of the GNU Lesser General Public
/// License along with this library; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#ifndef X3DTK_SFTYPE_H
#define X3DTK_SFTYPE_H

#include "X3DTypes.h"
#include "SFSceneGraph.h"
#include "SFComponent.h"

namespace X3DTK {

class X3DActor;

namespace X3D {
  class Creator;
  class X3DComponentCreator;
}

/*! \brief Class declared in <b><X3DTK/kernel.h></b> storing the informations relative 
 *  to the nodes of the derivation tree. 
 *
 *  It is based upon one singleton per tag. It means that it 
 *  exists only one instance of SFType of tag name "Cylinder" for the component "Geometry3D"
 *  of the "GL" scene graph.
 *
 *  \ingroup abstractNode
 */

class SFType
{
  friend class X3D::Creator;
  friend class Walker;
  friend class Visitor;
  friend class X3D::X3DComponentCreator;
  friend class X3DComponentWalker;
  friend class X3DComponentVisitor;
  friend class X3DAbstractNode;

public:
  /// Gets the SFType of name name.
  static SFType *getTypeOfName(const SFString &sceneGraph, const SFString &component, const SFString &name);
  /// Gets the SFType of id id.
  static SFType *getTypeOfId(int id);
  /// Recomputes the id and changes all the nodeProxy array.(Not yet implemented)
  static void recomputeIds();
  
  /// Returns the name of the node
  inline SFString getName() const {return _name;};
  /// Returns the name of the component of the node.
  inline SFString getComponentName() const {return _component->getName();};
  /// Returns the name of the component of the node.
  inline SFString getSceneGraphName() const {return _component->getSceneGraph()->getName();};
  /// Return the id of the node.
  inline int getId() const {return _id;};
  /// Gets the parent.
  inline SFType *getParent() const {return _parent;};
  /// Gets the list of the children.
  inline MFType getChildList() const {return _childList;};
  /// Gets the component.
  inline SFComponent *getComponent() const {return _component;};
  /// Gets the sceneGraph.
  inline SFSceneGraph *getSceneGraph() const {return _component->getSceneGraph();};
  /// Prints the current derivation tree.
  static void printDerivationTree();
    
private:
  /// Constructor.
  SFType(const SFString &name, const SFString &component, const SFString &sceneGraph);
  ~SFType();

  /// Sets the parent of the derivation hierarchy.
  void setParent(SFType *parent);
  /// Adds one instance of the SFType.
  void addOneReference();
  /// Adds one instance of the SFType and of the parent SFTypes.
  void addOneReferenceRecursive();
  /// Removes one instance of the SFType.
  static void removeOneReference(SFType *type);
  
  /// Defines a new Tag name. Only X3DNode is able to call this method.
  static void defineTypeName(SFType **type, const SFString &name, const SFString &component, const SFString &sceneGraph);
  /// Adds a new proxy. Only X3DNodeProxy is able to call this method.
  static void addActor(X3DActor *actor);
  /// Removes proxy from the list. Only X3DNodeProxy is able to call this method.
  static void removeActor(X3DActor *actor);

  /// Specifies to SFType that the next new Nodes are necessary to get their name and 
  /// will not be used then.
  static void beginFunctionDefinition();
  /// Specifies to SFType that the function definition is terminated.
  static void endFunctionDefinition();
  /// Computes the key from the sceneGraph, SFComponent and SFType strings.
  static SFString encodeKey(const SFString &sceneGraph, const SFString &component, const SFString &SFType);
  /// Returns the encoded name of the node
  inline SFString getEncodedName() const {return encodeKey(getSceneGraphName(), getComponentName(), getName());};
  
  /// attributes
  SFString _name;
  int _id;
  int _refCount;
  SFType *_parent;
  MFType _childList;
  SFComponent *_component;
  
  /// static attributes
  static int count;
  static MFType _typeList;  
  static std::list<X3DActor *> _actorList;
  static bool _functionDefinition;
 
  void printDerivationTree(SFString tab) const;
};

}

#endif
