////////////////////////////////////////////////////////////////////////////////
//                            X3D_MeshBuilderGeometry3DVisitor.inl            //
////////////////////////////////////////////////////////////////////////////////

namespace X3DTK {
namespace X3D {

template<class MData, class VData, class EData, class FData, bool readOnly>
MeshBuilderGeometry3DVisitor<MData, VData, EData, FData, readOnly>::MeshBuilderGeometry3DVisitor()
: Geometry3DVisitor()
{
  defineEnterFunction(&MeshBuilderGeometry3DVisitor::enterIndexedFaceSet);
  defineWalkOnFunction(&MeshBuilderGeometry3DVisitor::walkOnIndexedFaceSet);
  defineLeaveFunction(&MeshBuilderGeometry3DVisitor::leaveIndexedFaceSet);

  stateVariables = GraphTraversal::getInstanceOf<MeshBuilderStateVariables>();
}

template<class MData, class VData, class EData, class FData, bool readOnly>
MeshBuilderGeometry3DVisitor<MData, VData, EData, FData, readOnly>::~MeshBuilderGeometry3DVisitor()
{
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void MeshBuilderGeometry3DVisitor<MData, VData, EData, FData, readOnly>::enterIndexedFaceSet(IndexedFaceSet *I) const
{
  MESH::Mesh<MData, VData, EData, FData, readOnly> *TM = dynamic_cast<MESH::Mesh<MData, VData, EData, FData, readOnly> *>(stateVariables->getNode(I));
  if (TM == 0)
  {
    TM = new MESH::Mesh<MData, VData, EData, FData, readOnly>();
    TM->setName(I->getName());
    
    // normals
    if (MData::template find<MESH::MeshNormalData>())
    {
      TM->template ogetData<MESH::MeshNormalData>().setNormal(I->getNormal() != 0);
      TM->template ogetData<MESH::MeshNormalData>().setNormalPerVertex(I->getNormalPerVertex());
    }
    
    // colors
    if (MData::template find<MESH::MeshColorData>())
    {
      TM->template ogetData<MESH::MeshColorData>().setColor(I->getColor() != 0);
      TM->template ogetData<MESH::MeshColorData>().setColorPerVertex(I->getColorPerVertex());
      TM->template ogetData<MESH::MeshColorData>().setRGBA(dynamic_cast<X3D::ColorRGBA *>(I->getColor()) != 0);
    }
    
    // texCoords
    if (MData::template find<MESH::MeshTexCoordData>())
    {
      TM->template ogetData<MESH::MeshTexCoordData>().setTexCoord(I->getTexCoord() != 0);
    }
    
    ////////////////////////////////////////////////////////////////////////////
    // Coordinate
    
    Coordinate *C = dynamic_cast<Coordinate *>(I->getCoord());
  
    MESH::Vertex<VData, EData, FData, readOnly> *MC = dynamic_cast<MESH::Vertex<VData, EData, FData, readOnly> *>(stateVariables->getNode(C));
    if (MC == 0)
    {
      MC = new MESH::Vertex<VData, EData, FData, readOnly>();
      MC->setName(C->getName());
      stateVariables->addCoupleNode(C, MC);

      // Filling the Vertices.
      const MFVec3f &point = C->getPoint();

      for (MFVec3f::const_iterator it = point.begin(); it != point.end(); ++it)
      {
        typename MESH::Mesh<MData, VData, EData, FData, readOnly>::SFVertex *V = MC->createVertex();
        if (VData::template find<MESH::VertexPointData>())
          V->template ogetData<MESH::VertexPointData>().setPoint(SFPoint3f(*it));
      }  
    }
    TM->setVertex(MC);
    
    ////////////////////////////////////////////////////////////////////////////
    // Creating faces
        
    // future optim? reserve the vertices and edges.
    SFBool ccw = I->getCcw();
  
    //iterating the faces
    const MFInt32 &coordIndex = I->getCoordIndex();
    MFInt32::const_iterator itCoordIndex = coordIndex.begin();
  
    while (itCoordIndex != coordIndex.end())
    {
      MFInt32 faceCoordIndex;
      while (*itCoordIndex != -1)
      {
        faceCoordIndex.push_back(*itCoordIndex);
        ++itCoordIndex;
      }
      TM->createFace(faceCoordIndex, ccw);
      
      ++itCoordIndex;
    }
    
    ////////////////////////////////////////////////////////////////////////////
    // Other fields
    
    ////////////////// arrays //////////////////////////////////////////////////////////
  
    MFVec3f pointArray;
    MFVec3f normalArray;
    MFColorRGBA colorArray;
    MFVec2f texCoordArray;

    X3D::Normal *NormalNode = dynamic_cast<X3D::Normal *>(I->getNormal());
    if (NormalNode != 0)
      normalArray = NormalNode->getVector();
  
    if (I->getColor() != 0)
    {
      if (I->getColor()->getTypeName() == "ColorRGBA")
        colorArray = dynamic_cast<X3D::ColorRGBA *>(I->getColor())->getColor();
      else  
        colorArray = dynamic_cast<X3D::Color *>(I->getColor())->getColor();
    }
  
    X3D::TextureCoordinate *TexCoordNode = dynamic_cast<X3D::TextureCoordinate *>(I->getTexCoord());
    if (TexCoordNode != 0)
      texCoordArray = TexCoordNode->getPoint();
  
    bool hasNormal = !normalArray.empty();
    bool hasColor = !colorArray.empty();
    bool hasTexCoord = !texCoordArray.empty();
    
    ////////////////// indexes /////////////////////////////////////////////////////////
 
    const MFInt32 &colorIndex = I->getColorIndex();
    const MFInt32 &normalIndex = I->getNormalIndex();
    const MFInt32 &texCoordIndex = I->getTexCoordIndex();

    bool hasColorIndex = !colorIndex.empty();
    bool hasNormalIndex = !normalIndex.empty();
    bool hasTexCoordIndex = !texCoordIndex.empty();
    
    ////////////////// iterators ///////////////////////////////////////////////////////
    
    itCoordIndex = coordIndex.begin();
    MFInt32::const_iterator itColorIndex;
    MFInt32::const_iterator itNormalIndex;
    MFInt32::const_iterator itTexCoordIndex;
    unsigned int itFaceIndex = 0; 
    
    if (hasColorIndex)
      itColorIndex = colorIndex.begin();
    else
      itColorIndex = coordIndex.begin();
  
    if (hasNormalIndex)
      itNormalIndex = normalIndex.begin();  
    else
      itNormalIndex = coordIndex.begin();
  
    if (hasTexCoordIndex)
      itTexCoordIndex = texCoordIndex.begin();
    else
      itTexCoordIndex = coordIndex.begin();
    
    typename MESH::Mesh<MData, VData, EData, FData, readOnly>::MFFace::const_iterator itFace = TM->getFaces().begin(); 
    
    ////////////////// loop ////////////////////////////////////////////////////////////
  
    SFBool normalPerVertex = I->getNormalPerVertex();
    SFBool colorPerVertex = I->getColorPerVertex();
        
    while (itCoordIndex != coordIndex.end())
    {
      while ((*itCoordIndex != -1) && (itCoordIndex != coordIndex.end()))
      {
        MESH::SFVertex<VData, EData, FData, readOnly> *V = TM->getVertices()[*itCoordIndex];
     
        if (VData::template find<MESH::VertexColorData>() && hasColor)
          if (colorPerVertex)
            V->template ogetData<MESH::VertexColorData>().setColorOfFace(colorArray[*itColorIndex], *itFace);
          else
            (*itFace)->template ogetData<MESH::FaceColorData>().setColor(colorArray[(hasColorIndex ? *itColorIndex : itFaceIndex)]);  
              
        if (VData::template find<MESH::VertexNormalData>() && hasNormal)
          if (normalPerVertex)
            V->template ogetData<MESH::VertexNormalData>().setNormalOfFace(normalArray[*itNormalIndex], *itFace);
          else
            (*itFace)->template ogetData<MESH::FaceNormalData>().setNormal(normalArray[(hasNormalIndex ? *itNormalIndex : itFaceIndex)]);
            
        if (VData::template find<MESH::VertexTexCoordData>() && hasTexCoord)
          V->template ogetData<MESH::VertexTexCoordData>().setTexCoordOfFace(texCoordArray[*itTexCoordIndex], *itFace);  
      
        ++itCoordIndex;
        if (colorPerVertex)
          ++itColorIndex;     
        if (normalPerVertex)
          ++itNormalIndex;      
        ++itTexCoordIndex;
      }
    
      ++itCoordIndex;
      ++itColorIndex;
      ++itNormalIndex;
      ++itTexCoordIndex;
    
      ++itFace;
      ++itFaceIndex;
    }
    
    stateVariables->addCoupleNode(I, TM);
  }
  stateVariables->pushNode(TM);
}

template<class MData, class VData, class EData, class FData, bool readOnly>
bool MeshBuilderGeometry3DVisitor<MData, VData, EData, FData, readOnly>::walkOnIndexedFaceSet(IndexedFaceSet *, SFAbstractNode) const
{
  return false;
}

template<class MData, class VData, class EData, class FData, bool readOnly>
void MeshBuilderGeometry3DVisitor<MData, VData, EData, FData, readOnly>::leaveIndexedFaceSet(IndexedFaceSet *) const
{
  stateVariables->popNode();
}

}
}  
