/// This file is part of the X3DToolKit library
/// Copyright (C) 2002-2004  Yannick Le Goc (legoc@imag.fr)

/// http://artis.imag.fr/Members/Yannick.Legoc/X3D/

/// This library is free software; you can redistribute it and/or
/// modify it under the terms of the GNU Lesser General Public
/// License as published by the Free Software Foundation; either
/// version 2.1 of the License, or (at your option) any later version.

/// This library is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/// Lesser General Public License for more details.

/// You should have received a copy of the GNU Lesser General Public
/// License along with this library; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#ifndef X3DTK_X3D_X3DCOMPOSEDGEOMETRYNODE_H
#define X3DTK_X3D_X3DCOMPOSEDGEOMETRYNODE_H

#include "X3DTypes.h"
#include "X3D_X3DGeometry3DNode.h"

namespace X3DTK {
namespace X3D {

class X3DColorNode;
class X3DCoordinateNode;
class X3DNormalNode;
class X3DTextureCoordinateNode;

/*! \brief Base class declared in <b><X3DTK/X3D/scenegraph.h></b>, part of the <a href="../../doc/X3DSpecs/Part01/components/rendering.html">
 *  Rendering component</a> defining an X3DComposedGeometryNode
 *  (<a href="../../doc/X3DSpecs/Part01/components/rendering.html#X3DComposedGeometryNode">
 *  documentation</a>).
 *
 *  \ingroup X3DSceneGraph
 */

class X3DComposedGeometryNode : public X3DGeometry3DNode
{
public:
  /// Constructor.
  X3DComposedGeometryNode();
  /// Constructs an X3DComposedGeometryNode from its attributes.
  X3DComposedGeometryNode(X3DColorNode *color,
                          X3DCoordinateNode *coord,
                          X3DNormalNode *normal,
                          X3DTextureCoordinateNode *texCoord,
                          const SFBool &ccw,
                          const SFBool &colorPerVertex,
                          const SFBool &normalPerVertex,
                          const SFBool &solid);
  /// Virtual destructor.
  virtual ~X3DComposedGeometryNode();
 
  /// Sets the Color attribute.
  void setColor(X3DColorNode *color);
  /// Sets the Coord attribute.
  void setCoord(X3DCoordinateNode *coord);
  /// Sets the Normal attribute.
  void setNormal(X3DNormalNode *normal);
  /// Sets the TexCoord attribute.
  void setTexCoord(X3DTextureCoordinateNode *texCoord);
  /// Sets the Ccw attribute.
  void setCcw(const SFBool &ccw);
  /// Sets the ColorPerVertex attribute.
  void setColorPerVertex(const SFBool &colorPerVertex);
  /// Sets the NormalPerVertex attribute.
  void setNormalPerVertex(const SFBool &normalPerVertex);
  /// Sets the Solid attribute.
  void setSolid(const SFBool &solid);

  /// Gets the Color attribute.
  inline SFNode getColor() const {return _color;};
  /// Gets the Coord attribute.
  inline SFNode getCoord() const {return _coord;};
  /// Gets the Normal attribute.
  inline SFNode getNormal() const {return _normal;};
  /// Gets the TexCoord attribute
  inline SFNode getTexCoord() const {return _texCoord;};
  /// Gets the Ccw attribute.
  inline const SFBool &getCcw() const {return _ccw;};
  /// Gets the ColorPerVertex attribute.
  inline const SFBool &getColorPerVertex() const {return _colorPerVertex;};
  /// Gets the NormalPerVertex attribute.
  inline const SFBool &getNormalPerVertex() const {return _normalPerVertex;};
  /// Gets the Solid attribute.
  inline const SFBool &getSolid() const {return _solid;};
  
  /// Gets the children list.
  virtual MFAbstractNode getChildrenList() const;
  
  /// Adds the child N. Returns FALSE if the child already exists or N is not a 
  /// valid child.   
  virtual bool addChild(const SFAbstractNode &N);
  /// Sets the child N. Returns FALSE if N is not a 
  /// valid child. If the child already exists then N becomes the new child.
  virtual bool setChild(const SFAbstractNode &N);
  /// Removes the child if exists. Returns TRUE if succeeds.
  virtual bool removeChild(const SFAbstractNode &N);
  
  /// Loads the attributes from a X3DFileElement *.
  virtual void load(const X3DFileElement *element);
  /// Writes the attributes of the node.
  virtual SFString &write(SFString &output) const; 
 
protected:
  /// Copy constructor.
  X3DComposedGeometryNode(const X3DComposedGeometryNode &N);

private:  
  /// Color attribute.
  SFNode _color;  
  /// Coord attribute.
  SFNode _coord;  
  /// Normal attribute.
  SFNode _normal;
  /// TexCoord attribute.
  SFNode _texCoord;
  /// Ccw attribute.
  SFBool _ccw;
  /// ColorPerVertex attribute.
  SFBool _colorPerVertex;
  /// NormalPerVertex attribute.
  SFBool _normalPerVertex;
  /// Solid attribute.
  SFBool _solid;
};

}
}

#endif
